/* Kadohito Ohsuga*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#include <sys/poll.h>
#endif
#ifdef HAVE_SELECT_H
#include <sys/select.h>
#endif
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <strings.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <assert.h>
#include <errno.h>
#include <sys/mman.h>

#include "wcol.h"
#include "socket.h"
#include "sigs.h"
#include "msg.h"
#include "sim_stat.h"
#include "misc.h"
#include "http.h"
#include "sess.h"
#include "dhr.h"

extern int n_proxy;
extern int proxy_id;
extern PROXY_ID *cluster;

typedef union _ID{
    int id;
    char c[4];
}ID;

int dhr_hash1(char *host, char *object, int n){

  int i;
  register int hash_value = 0;
  register char *tmp;
  char ctoi;

  if((host == NULL) || (object == NULL))
    return -1;

  tmp = host;

  while(*tmp != '\0'){
    hash_value += *tmp;
    tmp++;
  }
 
  tmp = object;

  while(*tmp != '\0'){
    hash_value += *tmp;
    tmp++;
  }

  return hash_value % n;
}

int dhr_hash2(char *host, char *object, int n){

  int i;
  int hash_value1 = dhr_hash1(host, object, n);
  register int hash_value2 = 0;
  register char *tmp;
  char ctoi;

  if((host == NULL) || (object == NULL) || (hash_value1 < 0))
    return -1;

  tmp = host;

  while(*tmp != '\0'){
    hash_value2 += *tmp;
    tmp++;
  }
 
  tmp = object;

  while(*tmp != '\0'){
    hash_value2 += *tmp;
    tmp++;
  }

  if((hash_value2 % (n - 1)) >= hash_value1)
    return hash_value2 % (n - 1) + 1;
  else
    return hash_value2 % (n - 1);
 
}

int
dhr_proxy_failure(sess_t *sessp){

    char s;

#if 0
Trace("dhr_proxy_failure: state %s sess %d sfd %d server %d %s or %d %s\n"
       ,sessp->dhr_state_history
       ,sessp->sid
       ,sessp->sfd
       ,sessp->dhr_hash
       ,cluster[sessp->dhr_hash].address
       ,sessp->dhr_hash_tmp
       ,cluster[sessp->dhr_hash_tmp].address);
#endif
    DoHalfServerCloseSess(sessp);
    if(sessp->dhr_state == 'A'){
      sessp->dhr_hash_tmp = dhr_hash2(sessp->servername,
                                    sessp->serverpath, n_proxy);
      dhr_hash_swap(&(sessp->dhr_hash), &(sessp->dhr_hash_tmp));
      sessp->dhr_state = 'S';
    }
    else if(sessp->dhr_state == 'B'){
      sessp->dhr_hash_tmp = dhr_hash1(sessp->servername,
                                    sessp->serverpath, n_proxy);
      dhr_hash_swap(&(sessp->dhr_hash), &(sessp->dhr_hash_tmp));
      sessp->dhr_state = 'S';
    }
    else if(sessp->dhr_state == 'M'){
      sessp->dhr_hash_tmp = -1;
      dhr_hash_swap(&(sessp->dhr_hash), &(sessp->dhr_hash_tmp));
      s = 'S';
      dhr_add_history(sessp->dhr_state_history, s);
      sessp->dhr_state = 'E';
    }else{
      s = 'S';
      dhr_add_history(sessp->dhr_state_history, s);
      return 0;
    }

    sessp->state = STATE_QUERY_ADDR;
    return 1;

}

int
dhr_client_writev_insert(sess_t *sessp, struct iovec *iovec, int hlen, 
                         int nr, int size){

    int i;
    int w;
    int n_iov = 0;
    struct iovec iov[128];
    char dhr_header[64];

    iov[0].iov_base = iovec[0].iov_base;
    iov[0].iov_len  = hlen - 1;
    ++n_iov;

    dhr_header_insert(sessp->ishit, dhr_header);

    iov[1].iov_base = dhr_header;
    sessp->dhr_header_len = strlen(dhr_header);
    iov[1].iov_len = sessp->dhr_header_len;
    ++n_iov;
    iov[2].iov_base = iovec[0].iov_base + hlen - 1;
    iov[2].iov_len = nr - hlen + 1;
    ++n_iov;

    for(i=1; i<size; ++i){
        iov[i+2].iov_base = iovec[i].iov_base;
        iov[i+2].iov_len = iovec[i].iov_len;
        ++n_iov;
    }
        
    w = writev(sessp->cfd, iov, n_iov);

    return w - sessp->dhr_header_len;
}

void
dhr_header_insert(int ishit, char *header){

    switch(ishit){
      case SESS_MISS:
        sprintf(header,"DHRstate: %s MISS\r\n" ,cluster[proxy_id].address);
        break;
      case SESS_HIT:
        sprintf(header,"DHRstate: %s HIT\r\n" ,cluster[proxy_id].address);
        break;
      case SESS_DONT_TASK:
        sprintf(header,"DHRstate: %s DONT\r\n" ,cluster[proxy_id].address);
        break;
      default:
        sprintf(header,"DHRstate: %s -\r\n" ,cluster[proxy_id].address);
        break;
    }
}

int
dhr_header_check(sess_t *sessp, void *header, int hlen){

    int count = 0;
    int l;
    int i = 0;
    ID p_id;
    char dhr_header_element[16];
    char *p;

    p = header;

    while(i<hlen){
        if(*p=='D' && !strncmp(p,"DHRstate:",9)){
            l = 0;
            p += 10;
            i += 10;
            while(1){
                if(*p == ' '){
                    dhr_header_element[l] = '\0';
                    p++;
                    i++;
                    break;
                }
                else{
                    dhr_header_element[l++] = *p++;
                    i++; 
                }
            }
            for(p_id.id=0; p_id.id<n_proxy; ++p_id.id){
                if(!strcmp(cluster[p_id.id].address,dhr_header_element)){
                    p_id.id += 48;
                    sessp->judge[count*2] = p_id.c[0];
                    break;
                }
            }
            l = 0;
            
            if(!strncmp(p,"HIT",3)){
                sessp->judge[count*2+1] = 'h';
                p += 4;
                i += 4;
            }
            else if(!strncmp(p,"MISS",4)){
                sessp->judge[count*2+1] = 'm';
                p += 5;
                i += 5;
            }
            else if(!strncmp(p,"DONT",4)){
                sessp->judge[count*2+1] = 'd';
                p += 5;
                i += 5;
            }
            
            count++;
        }
        p++;
        i++;
    }

    return count;
}

int
dhr_state_change(sess_t *this){

    switch(this->dhr_state){
      int d1, d2;
    case 'L':
      d1 = dhr_hash1(this->servername, this->serverpath, n_proxy);
      d2 = dhr_hash2(this->servername, this->serverpath, n_proxy);
      if((d1 == proxy_id) || (d2 == proxy_id)){
        if(d1 == proxy_id){
          this->dhr_hash = d1;
          this->dhr_hash_tmp = d2;
          this->dhr_state = 'A';
        }
        else{
          this->dhr_hash = d2;
          this->dhr_hash_tmp = d1;
          this->dhr_state = 'B';
        }
      }
      else{
        this->cachable = SESS_UNCACHABLE;
        this->state = STATE_QUERY_ADDR;
        this->ishit = SESS_DONT_TASK;
        if(random()%2){
          this->dhr_hash = d1;
          this->dhr_state = 'A';
        }
        else{
          this->dhr_hash = d2;
          this->dhr_state = 'B';
        }
        return 0;
      }
      return 1;
    case 'A':
      this->dhr_hash = dhr_hash2(this->servername, this->serverpath, n_proxy);
      this->dhr_state = 'M';
      break;
    case 'B':
      this->dhr_hash = dhr_hash1(this->servername, this->serverpath, n_proxy);
      this->dhr_state = 'M';
      break;
    case 'S':
      this->dhr_state = 'E';
      this->dhr_hash_tmp = -1;
      dhr_hash_swap(&(this->dhr_hash),&(this->dhr_hash_tmp));
      break;
    case 'M':
      this->rcode = 306;
      this->dhr_hash = -1;
      this->dhr_state = 'R';
      break;
    case 'E':
      this->dhr_hash_tmp = -1;
      dhr_hash_swap(&(this->dhr_hash),&(this->dhr_hash_tmp));
      break;
    }

    return 1;

}

void
dhr_add_history(char *history, char dhr_state){

    int i;

    for(i=0; i<7; ++i){
        if(history[i] == '\0'){
            history[i] = dhr_state;
            history[i+1] = '\0';
            break;
        }
    }

}

void
dhr_hash_swap(int *h1, int *h2){
    
    int tmp;

    tmp = *h1;
    *h1 = *h2;
    *h2 = tmp;

}
    
