/*
 * KOTETU: the prefetching WWW proxy server
 *
 * disk_bench.c : benchmarking program for KOTETU
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1999, 2000
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#include <sys/poll.h>
#endif
#ifdef HAVE_SELECT_H
#include <sys/select.h>
#endif
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <strings.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <assert.h>
#include <errno.h>

/* private headers */

#include "range.h"
#include "msg.h"

#include "socket.h"
#include "sigs.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "misc.h"
#include "http.h"
#include "xsk.h"
#include "sess.h"
#include "squeue.h"
#include "parse.h"

#include "reason.h"

#include "mlist.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DEFAULT_CONFIG_ENVNAME
#define DEFAULT_CONFIG_ENVNAME "KOTETU"
#endif

#ifndef DEFAULT_CONFIGFILE
#define DEFAULT_CONFIGFILE      "/etc/kotetu.conf"
#endif
#ifndef DEFAULT_LOGFILE
#define DEFAULT_LOGFILE         "kotetu.log"
#endif

#ifndef MONITOR_INTERVAL
#define MONITOR_INTERVAL    (600)
#endif

#ifndef MALLOC
#define MALLOC(x)   malloc(x)
#endif
#ifndef FREE
#define FREE(x)     free(x)
#endif

#define KOTETU_VER  "Kotetu-1.0"


/* dummy */

#define cache_show_blockarray(x,y)      ((void)0)
#define cache_list_entry(x)             ((void)0)
int nline;
int nblock;
int linesize;



pthread_mutex_t stat_lock = PTHREAD_MUTEX_INITIALIZER;
time_t boot_date = 0;
time_t service_start_date = 0;
int    session_queue_report_flag = 0;

volatile long long request_sirial_count = 0;
volatile int receive_request_count = 0;
volatile int process_request_count = 0;
volatile int issue_prefetch_count  = 0;
volatile int report_request_count  = 0;
volatile int reject_request_count  = 0;
volatile int ignore_request_count  = 0;

volatile int send_reply_count    = 0;
volatile int hit_reply_count     = 0;
volatile int miss_reply_count    = 0;
volatile int error_reply_count   = 0;

volatile int hash_colision      = 0;
volatile int compaction_count   = 0;
volatile int msweeper_count     = 0;
volatile int expire_count       = 0;

volatile int skip_cache_count   = 0;
volatile int alloc_tmp_buf_count = 0;
volatile int alloc_sess_buf_count = 0;

extern volatile int swapload_count;
extern volatile int swapin_count;
extern volatile int swapout_count;
extern volatile int swapload_success;
extern volatile int swapin_success;
extern volatile int swapout_success;

extern volatile int swapout_token;
extern volatile int swapout_token_init;
extern volatile int swapout_token_interval;

extern volatile int dsweeper_seq;
extern volatile int dsweeper_count;

int read_reply_total   = 0;
int read_reply_success = 0;
int read_reply_fail    = 0;

int cache_newover_count = 0;
int cache_newfree_count = 0;
int cache_newmin_count = 0;



volatile off_t intra_incoming = (off_t)0; /* requests from clients */
volatile off_t extra_outgoing = (off_t)0; /* requests to super class servers */
volatile off_t extra_incoming = (off_t)0; /* replys from super class servers */
volatile off_t intra_outgoing = (off_t)0; /* replys to clients */



sim_stat_rec rcode_stat;
long    rcode_stat_count[1000];

sim_stat_rec ret_time;
long    ret_time_count[K(7)];

sim_stat_rec ret_time_short;
long    ret_time_short_count[K(32)];

sim_stat_rec len_perread_stat;
static long server_buffer_stat_count[K(32)];

sim_stat_rec object_size_stat;
static long object_size_stat_count[K(1)];

sim_stat_rec url_first_hash_stat;
static long url_first_hash_stat_count[K(2)];
sim_stat_rec url_second_hash_stat;
static long url_second_hash_stat_count[K(2)];
sim_stat_rec url_third_hash_stat;
static long url_third_hash_stat_count[K(10)];
sim_stat_rec url_fourth_hash_stat;
static long url_fourth_hash_stat_count[K(10)];

sim_stat_rec req_field_type_stat;
static long req_field_type_stat_count[K(10)];

sim_stat_rec object_alloc_stat;
static long object_alloc_stat_count[K(10)];

sim_stat_rec mark_time;
static long mark_time_count[K(10)];
sim_stat_rec msweep_time_single;
static long msweep_time_single_count[K(10)];
sim_stat_rec msweep_time_long;
static long msweep_time_long_count[K(10)];


pthread_t th_listener;
pthread_mutex_t accept_lock = PTHREAD_MUTEX_INITIALIZER;

pthread_t th_handlers[N_HANDLER_MAX];
pthread_t th_monitor;

pthread_t th_signalhandler;

sigset_t handler_sig_mask;

int require_sess_length;

int think_queue_length;
sess_queue_t X_think_queue;
sess_queue_t *think_queue;
#if 0
sess_queue_t X_send_queue;
sess_queue_t *send_queue;
sess_queue_t X_fetch_queue;
sess_queue_t *fetch_queue;

sess_queue_t X_prompt_queue;
sess_queue_t *prompt_queue;
#endif

sess_queue_t X_wait_queue;
sess_queue_t *wait_queue;

sess_queue_t X_msweep_queue;
sess_queue_t *msweep_queue;

#ifndef N_SWEEPER_MAX
#define N_SWEEPER_MAX N_HANDLER_MAX
#endif
sess_queue_t X_swapout_queue[N_SWEEPER_MAX];
sess_queue_t *swapout_queue[N_SWEEPER_MAX];

#if 0
sess_queue_t X_swapout_queue;
sess_queue_t *swapout_queue;
#endif
sess_queue_t X_swapin_queue;
sess_queue_t *swapin_queue;


#ifdef CHECK_QUEUE

#define push_sess(q,c) \
    RawTrace("push sess# %d, %p to %p : %d/%d : %s %5d\n",\
        c->sid, c, q, q->count, q->cmax, __FILE__ , __LINE__), sq_push(q,c)
#define pull_sess(q) \
    RawTrace("pull from %p : %s %5d\n", q, __FILE__ , __LINE__), sq_pull(q,c)

#else
#define push_sess(q,c) sq_push(q,c)
#define pull_sess(q) sq_pull(q)
#endif
#define pull_sess_nonblock(q) sq_pull_nonblock(q)



sim_hist_rec comp_hist;
static long comp_hist_area[128];
sim_hist_rec expire_hist;
static long expire_hist_area[128];

extern sim_hist_rec fratio_hist;

/*
 * externals
 */

extern int nhot;

extern int radix_h1, radix_h2, radix_h3, radix_h4;

#if 0
#define push_sess(q,c) sq_timedpush(q,c,100000) /* 100 msec */
#endif


int InitSession(int);


/* automatic */
pthread_mutex_t config_lock = PTHREAD_MUTEX_INITIALIZER;

char config_filename[L_PATH];
char *log_filename;
int trace_flag;
int monitor_flag;
int prefetch_flag=1;
int service_portno;
int require_n_handler;
int require_n_dsweeper;
int require_n_msweeper;
int require_chunk_size;
off_t require_total_memory_size;

int require_num_bucket;
int require_level_bucket_skiplist;
int require_size_frag;
int require_ratio_frag_length;

int num_prefetch_referedpages;
int num_prefetch_includedobjs;

int nfrag_obj       = HARD_NFRAG_OBJ;
int nfrag_nwrite    = HARD_NFRAG_NWRITE;
int nfrag_nread     = HARD_NFRAG_NREAD;
int nfrag_dwrite    = HARD_NFRAG_DWRITE;
int nfrag_dread     = HARD_NFRAG_DREAD;

int require_nfrag_obj;
int require_nfrag_nwrite;
int require_nfrag_nread;
int require_nfrag_dwrite;
int require_nfrag_dread;

int last_keep_depth;
int cache_normal_idle_th_base   = 3600;
int cache_normal_idle_th        = 3600;
int cache_cracked_idle_th_base  =  600;
int cache_cracked_idle_th       =  600;


/*
pthread_mutex_t gc_lock = PTHREAD_MUTEX_INITIALIZER;
*/


char *localdomain_padding;
char *report_hook_hostname;

/* static */
pid_t    wcol_pid;
char     *progname;


extern off_t sess_memory_size;
off_t cache_memory_size;


extern int sq_check_count();
extern int check_emptyentity(bucket_t* bk);


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, msleep is made.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;

    w.tv_sec = 0;
    w.tv_nsec = x * 1000000;   /* convert msec to nsec */

    nanosleep(&w, NULL);

    return 0;
}
#endif



int
check_sq(char *msg)
{
    Trace("# %s: sq_count(free) %d, diff %d, think Q %d/%d, wait Q %d/%d, msweep Q %d/%d\n",
        msg,
        sq_check_count(),
        sq_check_count()
            - (think_queue->count + wait_queue->count + msweep_queue->count),
        think_queue->count, think_queue->cmax,
        wait_queue->count, wait_queue->cmax,
        msweep_queue->count, msweep_queue->cmax);
    return 0;
}

int
CHKTIME(time_t thetime)
{
    int diff;
    diff = time(NULL) - thetime;
    if(diff<0) diff = -diff;
    if(diff>86400) {
        return 0;
    }
    else {
        return 1;
    }
}

#if 0
void
dump_bkinfo()
{
    int i;
    bucket_t *b;

    for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
        Trace(" %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
            i,
            b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
            b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
    }
}
#endif


/* free list of entry or frag is empty, when the number of use is smaller
 * than high water.
 */
int
check_bkinfo()
{
    int i;
    bucket_t *b;
    int r, c;

    c = 0;
    for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
        r = 0;
        if(b->entry_freelist==NULL && b->entry_use <  b->entry_hwater) {
            r++;
        }
#if 0
        if(b->frag_freelist==NULL && b->frag_use <  b->frag_hwater) {
            r++;
        }
#endif

        if(r) {
#if 0
          Trace(" be %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
            i,
            b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
            b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
#endif
          Trace(" bk# %d, entry u %4d f %d hw %d all %d, %08x, frag %4d\n",
            i,
            b->entry_use, b->entry_free, b->entry_hwater, b->entry_all,
            b->entry_freelist,
            b->frag_use);

            c++;
        }
    }

    if(c==0)
        return 1;
    else
        return 0;
}

pthread_mutex_t wait_check_lock;

inline int check_timeout(sess_t *sessp, time_t now);

int
check_friend_lock(sess_t *this)
{
    sess_t *cu;
    int n;
    int i;
    int chk;
    int w, d;
    time_t now;
    int idle;

    if(this==NULL || this->cache_entry==NULL) {
        return -1;
    }
Trace("check_friend_lock:\n");
Trace("  target sess# %d, oid# %d\n",
    this->sid, ((skcache_entry_t*)this->cache_entry)->oid);


    MUTEX_LOCK(&wait_check_lock);

    now = time(NULL);

    n = wait_queue->count*2; /* XXX */
    i = 0;
    w = 0;
    d = 0;
    while(i<n) {
        cu = pull_sess_nonblock(wait_queue);
        if(cu==NULL) {
            break;
        }
        idle = now - cu->last_action;

#if 0
  if(cu->cache_entry) {
    Trace("  %3d: sess# %d, %p %s : entry %p, oid# %d, ref %d, owner %p, idle %d\n",
    i,
    cu->sid, cu, MSG_STATE_[cu->state],
    cu->cache_entry,
    ((skcache_entry_t*)cu->cache_entry)->oid,
    ((skcache_entry_t*)cu->cache_entry)->ref,
    ((skcache_entry_t*)cu->cache_entry)->owner, idle
    );
  }
#endif

        if(check_timeout(cu, now)) {
            cu->state = STATE_QUERY_ADDR;
            cu->cachable = SESS_UNCACHABLE;

            chk = push_sess(think_queue, cu);
            if(chk) {
Trace("timeout waited friend sess# %d from wait Q, fail to push think Q\n",
                    cu->sid);
                d++;

                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);
                cu = NULL;

            }
            else {
Trace("timeout waited friend sess# %d from wait Q, push think Q\n",
                    cu->sid);
                w++;
            }
        }
        else
        if(cu->cache_entry &&
            ((skcache_entry_t*)cu->cache_entry)->oid ==
                ((skcache_entry_t*)this->cache_entry)->oid) {

            /* re-push to think Q */
            chk = push_sess(think_queue, cu);
            if(chk) {
Trace("pick up waited friend sess# %d from wait Q, fail to push think Q\n",
                    cu->sid);
                d++;

                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);
                cu = NULL;

            }
            else {
Trace("pick up waited friend sess# %d from wait Q, push think Q bk# %d,\n",
                    cu->sid, cu->fhash);
                w++;
            }
            cu = NULL;
        }
        else {
#if 0
            chk = push_sess(wait_queue, cu);
#endif
            chk = sq_timedpush(wait_queue, cu, 5000);
            if(chk) {
                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);

            }
            cu = NULL;
        }
        i++;
    }

    MUTEX_UNLOCK(&wait_check_lock);

    Trace("check_friend_lock: sess# %d, oid# %d, n %d, w %d, d %d\n",
        this->sid, ((skcache_entry_t*)this->cache_entry)->oid,
        n, w, d);

    return w;
}

/*
 * release_buf - release allocated area
 *
 */
int
release_buf(sess_t *this)
{
    int ret;
    int chk;
    bucket_t *bk;
    skcache_entry_t *centry;


#ifdef TRACE_CACHE
    Trace("release_buf: sess# %d, '%s' [%d:%d] %s\n",
        this->sid, this->loc, this->fhash, this->shash,
        MSG_STATE_[this->state]);

#if 0
    cache_showblockaddr(this->fhash, 0);
#endif

#endif

    bk = &skcache_bucket_base[this->fhash];

    LOCK_BK(bk);


    centry = this->cache_entry;
    if(centry) {
        centry->ref--;

#if 0
        if(centry->owner==this->sid) {
            centry->owner = 0;
        }
#endif
        if(centry->owner==this) {
            centry->owner = NULL;
        }

        if(centry->ref>0) {
#ifdef USE_WQUEUE
            check_friend_lock(this);
#endif
        }

        this->cache_entry = NULL;
    }

    ret = 0;

 assert(bk->cup.num==bk->entry_use);
    UNLOCK_BK(bk);
    bk = NULL;

out:

    return ret;
}





int
HostnameNormalize(char *src)
{
    register char *p;
    register int c;
    int d;

    p = src;
    c = 0;
    while(*p) {
        c++;
        if(*p=='.') {
            goto have_dot;
        }
        p++;
    }

not_have_dot:

    MUTEX_LOCK(&config_lock);
    if(c+strlen(localdomain_padding)+1<L_HOST) {
        strcat(p, localdomain_padding);
    }
    MUTEX_UNLOCK(&config_lock);

    return 0;

have_dot:
    return 0;
}


#if 0
int check_bkentry(bucket_t*,char*);
#endif


extern int skcache_mem_keep_min;
extern int skcache_mem_keep_cur;
extern int skcache_mem_keep_max;
extern int skcache_entry_keep_min;
extern int skcache_entry_keep_cur;
extern int skcache_entry_keep_max;


int
scan_frags_in_bucket(bucket_t *bk)
{
    skiplist_t *list;
    skiplist_cell_t *x;
    skcache_entry_t *ent;
    skcache_frag_t *cur;
    int in_entry;
    int in_bucket;

    list = &bk->cup;
    x = list->header->skip[0];

    in_bucket = 0;

    while(x!=list->term) {
        ent = (skcache_entry_t*) x->entity;
        cur = ent->m_recorded;

        in_entry = 0;

        while(cur) {
            if(cur->next==cur) {
                Error("bk#%d, entity %p, %d th frag %p is looped itself",
                    bk->bid, ent, in_entry, cur);
                ent->cracked = 1;
                break;
            }
            in_entry++;

            cur = cur->next;
        }
        in_bucket += in_entry;
        
        x = x->skip[0];
        if(x==NULL)
            break;
    }

#if 0
    Trace("bk# %d, has %d frag\n", bk->bid, in_bucket);
#endif

    return in_bucket;
}








/* support only IPv4 */
static
int
store_scrambled_name(char *dst, struct sockaddr_in *addr, int sc_key)
{
    register char *p, *q;
    register int i;
    register int c;

    q = (char*) &addr->sin_addr;

    c = 0xc634d3a5;
    c |= sc_key;
    for(i=3;i>=0;i--) {
        p = &q[i];
        c += *p;
        c = c << 3 + 0x01;

#if 0
        Trace(" #%d %3u %2x -> %12u %12x\n",
            i, (unsigned int)*p, (unsigned int)*p,
            (unsigned int)c, (unsigned int)c);
#endif
    }

    sprintf(dst, "c%u", c);

    return 0;
}

static
int
store_direct_name(char *dst, struct sockaddr_in *addr)
{
    register char *p, *q;
    register int i;
    register int c;

    p = (char*) &addr->sin_addr;
    q = dst;

    for(i=0;i<4;i++) {
        if(*p>100)
            *q++ = '0' + (*p / 100);
        if(*p>10)
            *q++ = '0' + ((*p % 100)/10);
        *q++ = '0' + (*p % 10);
        *q++ = '.';
        *q = '\0';

#if 0
        Trace(" #%d %3u %2x -> '%s'\n",
            i, (unsigned int)*p, (unsigned int)*p, dst);
#endif

        p++;
    }

    return 0;
}


static
int client_scramble_key;




inline
int
check_timeout(sess_t *sessp, time_t now)
{
    int ret;
    int idle, elap;

    ret = 0;

    idle = now - sessp->last_action;
    elap = now - sessp->accept_time.tv_sec;

    switch(sessp->state) {
    case STATE_ACCEPT:
    case STATE_RECV_REQ:
#if 0
    Trace("check_timeout: sess# %d, now %d, idle %d, elap %d, reqlen %d\n",
        sessp->sid, now, idle, elap, sessp->reqlen);
#endif
        if(sessp->last_action == sessp->accept_time.tv_sec) {

            if(sessp->reqlen>0) {
#if 0
    Trace("    Is it empty request ?\n");
#endif
            }

            if(idle > RECEPT_TIMEOUT*3) {
                sessp->reason = R_CLIENT_TIMEOUT;
                ret = 1;
            }
        }
        else
        if(idle > RECEPT_TIMEOUT) {
            sessp->reason = R_CLIENT_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_FRIEND:
        if(idle > WFRIEND_TIMEOUT) {
            sessp->reason = R_WFRIEND_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_CONN_SERV:
    case STATE_RECV_RES:
        if(idle > FETCH_TIMEOUT) {
            sessp->reason = R_SERVER_TIMEOUT;
            ret = 1;
        }
        break;
    default:
        if(idle > ELSE_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
        break;
    }

    if(ret==0) {
        if(elap > SESSION_TOTAL_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
    }

#if 0
if(ret!=0) {
    Trace("sess# %d, elap/idle %d/%d -> reason %d, ret %d\n",
        sessp->sid, elap, idle, sessp->reason, ret);
}
#endif

    return ret;
}



#if 0
#define CHECK {\
    int k;\
    Trace("w_sess_count %d, w_fd_count %d\n", w_sess_count, w_fd_count);\
    if(w_sess_count>0) {\
        Trace("index sess     sid#  state           |cfd   sfd\n");\
        for(k=0;k<w_sess_count;k++) {\
            if(sess_chunk[k]) \
                Trace("%5d %8x %5d %-15s |%5d %5d \n",\
                    k, sess_chunk[k],\
                    sess_chunk[k]->sid, MSG_STATE_[sess_chunk[k]->state],\
                    sess_chunk[k]->cfd, sess_chunk[k]->sfd\
                    );\
            else\
                Trace("%5d empty\n", k);\
        }\
    }\
    else {\
        Trace("***no sessions\n");\
    }\
    if(w_fd_count>0) {\
        Trace("\nindex fd    |sess   sid# state\n");\
        for(k=0;k<w_fd_count;k++) {\
            if(bound[k]) \
                Trace("%5d %5d |%8x %5d %-15s\n",\
                    k, fdrec[k].fd,\
                    bound[k], bound[k]->sid, MSG_STATE_[bound[k]->state]);\
            else\
                Trace("%5d %5d empty\n", k, fdrec[k].fd);\
        }\
    }\
    else {\
        Trace("***no fds\n");\
    }\
    Trace("= = \n");\
}

#else
#define CHECK   ((void)0)
#endif

static
int
sess_comp(const void *ap, const void *bp)
{
    sess_t *a = *(sess_t**)ap;
    sess_t *b = *(sess_t**)bp;

    if(a->fhash==b->fhash) {
        return a->shash - b->shash;
    }
    else {
        return a->fhash - b->fhash;
    }
}





extern volatile int disk_sweep_interval;

void*
DSweeper(void* dummy)
{
    int chk;
    struct timespec join_time;

    /* msec, how many times in 1000msec */
    int interval = disk_sweep_interval;

#if 1
Trace("DSweeper: pr %d, th %d\n", getpid(), pthread_self());
#endif

    while(1) {
        skcache_dir_sweeper();
        chk = msleep(interval);
    }
}

extern sim_stat_rec swapout_time;

#define DLEN    K(11)

aLoop(int xc, int xw, int xi)
{
    int i, j;
    bucket_t *bk;
    skcache_entry_t *newone;
    skcache_entry_t *stored_pos;
    skcache_frag_t *cur;
    skcache_frag_t *last;
    int h1, h2, h3, h4;
    int already_have;
    int have, nothave;
    struct timeval xbefore, xafter;
    struct timeval ybefore, yafter;
    int takes;
    int chk;
    long long sizesum;
    off_t olen;

    long long tmpsum;

    char *hfmt= "#%7s %6s %12s %7s %8s %7s %7s\n";
    char *fmt = "%8d %6d %12lld %7.1f %8d %7.2f %7.2f\n";
    
    char sample_data[2*DLEN];


    srand(time(NULL) & getpid());

    have = 0;
    nothave = 0;
    j = 0;
    
    sizesum = 0LL;
    tmpsum = 0LL;

    fprintf(stdout, "\n");
    fprintf(stdout, hfmt, "serial", "lap",
        "size","", "take", "rate", "");
    fprintf(stdout, hfmt, "-", "-",
        "B","GB", "ms", "op/s", "MB/s");

    gettimeofday(&xbefore, NULL);
    gettimeofday(&ybefore, NULL);

    for(i=0;i<xc;i++) {
        h1 = rand() % radix_h1;
        h2 = rand() % radix_h2;
        h3 = rand() % radix_h3;
        h4 = rand() % radix_h4;

        olen = DLEN + DLEN/2 - (rand()%DLEN);

        bk = &skcache_bucket_base[h1];

        LOCK_BK(bk);

        newone = skcache_pull_new_entry(bk);
        if(newone==NULL) {
            Error("no memory");
            break;
        }
        newone->oid = i;
        newone->h1 = h1;
        newone->h2 = h2;
        newone->h3 = h3;
        newone->h4 = h4;
        sprintf(newone->name, "someURL%d", i);

        stored_pos =
            (skcache_entry_t*) sk_assign(&bk->cup, newone,
                &bk->elem_freelist, skcache_entry_cmp,
                &already_have, 0);

        if(stored_pos==NULL || already_have<0) {
            Trace("have %d,%d,%d,%d\n", h1, h2, h3, h4);

            skcache_push_idle_entry(bk, newone);

            have++;
            continue;
        }
        else {
            nothave++;
        }


        stored_pos->m_recorded = NULL;
        stored_pos->d_recorded = -1;
        stored_pos->count = 0;          /* always new */

        {
            int c;
            off_t len;
            off_t alen;

            c = 0;
            len = 0;
            last = NULL;

            alen = olen;
    
            do {
                cur = skcache_pull_new_frag(bk);
                if(!cur) {
                    goto loop_end;
                }

                if(last==NULL) {
                    stored_pos->m_recorded = cur;
                }
                else {
                    last->next = cur;
                }

                memcpy(cur->ent, &sample_data[len], MIN(cur->size, alen));
                cur->len = MIN(cur->size, alen);
                cur->next = NULL;

                alen -= MIN(cur->size, alen);

                c++;
                len += (off_t)cur->len;
                last = cur;
            } while(alen>0);

#if 0
            Trace("store %ld bytes in %d frags\n", (long)len, c);
#endif
        }

        if(stored_pos->m_recorded==NULL) {
            Trace("stored_pos %p, m_recorded %p\n",
                stored_pos, stored_pos->m_recorded);
        }
        
        chk = skcache_entry_swapout(bk, stored_pos);
        if(chk) {
            Error("fail swapout <%d>", chk);
            if(chk==XSK_DISKFULL) {
                break;
            }
        }
        else {
            chk = skcache_entry_delete(bk, stored_pos);
        }

        j++;
        tmpsum += (long long)olen;
        sizesum += (long long)olen;

        UNLOCK_BK(bk);

        if(i%xw==xw-1) {
            gettimeofday(&yafter, NULL);
            takes = KOTETU_SUB_TIME(yafter, ybefore);
            fprintf(stdout, fmt, i, j,
                (long long)sizesum, fdG(sizesum),
                takes, ((double)xw/takes)*1000.0,
                ((double) (tmpsum)/takes/(1024*1024))*1000.0
                );
            fflush(stdout);

#if 0
Trace("\tfrag use %d, free %d\n", frag_use, frag_free); 
#endif

            if(xi)
                    msleep(xi);

            gettimeofday(&ybefore, NULL);
            j = 0;
            tmpsum = 0LL;
        }
    }
loop_end:
    gettimeofday(&yafter, NULL);
    gettimeofday(&xafter, NULL);

    if(j) {
            takes = KOTETU_SUB_TIME(yafter, ybefore);
            fprintf(stdout, fmt, i, j,
                (long long)sizesum, fdG(sizesum),
                takes, ((double)(i%xw)/takes)*1000.0,
                ((double) (tmpsum)/takes/(1024*1024))*1000.0
                );
    }

    takes = KOTETU_SUB_TIME(xafter, xbefore);

    fprintf(stdout, "\n");

    if(i>0) {
        fprintf(stdout, "#");
        fprintf(stdout, fmt, i, i,
            (long long)sizesum, fdG(sizesum),
            takes, ((double)i/takes)*1000.0);
        fprintf(stdout, "\n");
    }

    fflush(stderr);
    fflush(stdout);
#if 0
    sim_stat_show99_lock(&swapout_time, stdout);
#endif

}


void*
DiskJudge(void* dummy)
{
#if 0
    aLoop(50*20, 50, 0);
#endif
    aLoop(20000*500, 20000, 0);

    exit(0);
}





extern sim_stat_rec hres_stat;



extern long long swap_count;
extern long long swap_size;


extern sim_stat_rec swapload_time;
extern sim_stat_rec swapin_time;
extern sim_stat_rec swapout_time;
extern sim_stat_rec checkdir_time;
extern sim_stat_rec compaction_time;

extern sim_stat_rec swapout_fail_reason;
extern sim_stat_rec swapload_fail_reason;

extern int dsweeper_phase;


static
int
check_mutex(pthread_mutex_t *mx)
{
    int chk;

    chk = MUTEX_TRYLOCK(mx);
    if(chk) {
        if(chk==EBUSY)
            return 0;
        else
            return -chk;
    }
    MUTEX_UNLOCK(mx);
    return 1;
}


static char *_msg_idle  = "IDLE ";
static char *_msg_busy  = "BUSY ";
static char *_msg_err   = "ERROR";

static
int
report_check_mutex(FILE *fp, char *msg, pthread_mutex_t *mx)
{
    int chk;
    char *rep;

    chk = check_mutex(mx);
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    fprintf(fp, "$$$ %s %s\n", msg, rep);

    return 0;
}

void print_sess(FILE *ofp, void* raw_addr, struct timeval *reftime)
{
    sess_t *self;
    int chk, err;
    char *rep;
    struct timeval reftime_local;
    int elap;

    if(reftime==NULL) {
        gettimeofday(&reftime_local, NULL);
        reftime = &reftime_local;
    }

    self = (sess_t*) raw_addr;

    chk = check_mutex(&(self->busy));
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    elap = reftime->tv_sec - self->accept_time.tv_sec;

    fprintf(ofp, "%3d %s %c %3d %3d <%s>\n",
        self->sid, rep, self->mode==MODE_CLIENT ? 'c': '-', elap, self->fhash, self->loc);
#if 0
    fprintf(ofp, "%3d ", self->reqlen);
    fprintf(ofp, "<%s> ", self->loc);
    fprintf(ofp, "\n");
#endif
}

int
check_buckets()
{
    int i;
    bucket_t *bk;
    int f;

    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];
        f = check_mutex(&bk->mutex);
        Trace("bk# %4d: %d\n", i, f);
    }

    return 1;
}

int
check_busy_buckets()
{
    int i;
    bucket_t *bk;
    int f;
    int c;
    char band[RADIX_H1_MAX];
    char fill[RADIX_H1_MAX];

    c = 0;
    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];

/*
        fill[i] = "_ueX"[(bk->entry_use>0)*1+(bk->gc_checkpoint>0)*2];
*/
        fill[i] = "0123456789X"[10*bk->entry_use/bk->entry_all];

        f = check_mutex(&bk->mutex);
        if(f==0) {
#if 0
            if(c==0) {
                Trace("busy buckets:\n");
            }
#endif
#if 0
            Trace("busy bk# %-3d: %d %d\n", i, f, bk->gc_last);
#endif
            c++;
            if(bk->own=='S') {
                band[i] = bk->own;
            }
            else {
                band[i] = '*';
            }
        }
        else
        if(f<0) {
            band[i] = '?';
        }
        else {
            band[i] = '.';
        }
    }
    band[radix_h1] = '\0';
    fill[radix_h1] = '\0';

    if(c>0) {
#if 0
        Trace("busy %d threads\n", c);
#endif
#if 0
    Trace("busy band %d %3d %s\n", time(NULL), c, band);
#endif
    }
#if 1
    Trace("busy band %d %3d %s\n", time(NULL), c, band);
    Trace("busy fill %d %3d %s\n", time(NULL), c, fill);
#endif

    return 1;
}



static pthread_mutex_t logging_lock = PTHREAD_MUTEX_INITIALIZER;

RETSIGTYPE
Logging_Queues()
{
    FILE *fp;
    int i;
    struct timeval now;
    int t_nhot;

    MUTEX_LOCK(&logging_lock);

    gettimeofday(&now, NULL);

    Trace("logggin queues\n");

    fp = fopen("queues.log", "w");
    if(fp==NULL) {
        goto out;
    }

    fprintf(fp, "process %d, thread %x\n\n", getpid(), pthread_self());

        fprintf(fp, " monitor           %x\n", th_monitor);
        fprintf(fp, " listener          %x\n", th_listener);
    fprintf(fp, " manager  %d/%d\n", require_n_handler, N_HANDLER_MAX);
    for(i=0;i<require_n_handler;i++) {
        fprintf(fp, "   %3d             %x\n", i, th_handlers[i]);
    }
    fprintf(fp, "\n");

    MUTEX_LOCK(&sess_lock);
    t_nhot = nhot;
    MUTEX_UNLOCK(&sess_lock);

    fprintf(fp, "hot session %d\n\n", t_nhot);

    fprintf(fp, "== think queue\n");
    sq_show_mapfunc(fp, think_queue,    print_sess, &now);
#if 0
    fprintf(fp, "== send queue\n");
    sq_show_mapfunc(fp, send_queue,     print_sess, &now);
    fprintf(fp, "== fetch queue\n");
    sq_show_mapfunc(fp, fetch_queue,    print_sess, &now);
    fprintf(fp, "== prompt queue\n");
    sq_show_mapfunc(fp, prompt_queue,   print_sess, &now);
#endif


    report_check_mutex(fp, "mutex stat   ", &stat_lock);
    report_check_mutex(fp, "mutex sess   ", &sess_lock);
    report_check_mutex(fp, "mutex config ", &config_lock);
    report_check_mutex(fp, "mutex logging", &logging_lock);


done:

    fflush(fp);
    fclose(fp);

out:

    MUTEX_UNLOCK(&logging_lock);
}


/*
 *
 */
RETSIGTYPE
TERM_Trap()
{
    Log("; Catch signal for termination");
    MUTEX_LOCK(&stat_lock);
    catch_termination = 1;
    MUTEX_UNLOCK(&stat_lock);
    exit(127);
}

void*
signal_watcher(void *arg)
{
    int sig;
    int chk;

    while(1) {
        chk = sigwait(&handler_sig_mask, &sig);
        if(chk) {
            Trace("signal_watcher: chk %d\n", chk);
            continue;
        }

        Trace("signal_watcher: get signal %d\n", sig);

        switch(sig) {
        case SIGPIPE:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGUSR2:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGHUP:
            LogSync();
            break;
        case SIGINT:
        case SIGTERM:
            MUTEX_LOCK(&stat_lock);
            catch_termination = 1;
            MUTEX_UNLOCK(&stat_lock);
            Log("END");
            LogSync();
            exit(137);
            break;
        case SIGCHLD:
          {
            int stat;
            pid_t pid;

            do {
try_waitpid:
                pid = waitpid((pid_t)-1, &stat, WNOHANG | WUNTRACED);

                if(pid < 0) {
                    if(errno==EINTR) {
                        goto try_waitpid;
                    }
                    printf("error with errno=%d\n", errno);
                    break;
                }
                if (pid > 0) {
                    printf("child %d terminated\n", pid);
                }
            } while(pid>0);

          }
            break;
        default:
            break;
        }
    }

 /*NOTREACHED*/
    return (void*)NULL;
}



/*
 *
 * if fdcap is -1, this routine will set maximum value from getrlimit()
 */
void
ToBeMaximumFDCapacity(int fdcap)
{
    struct rlimit nfile;
    int last_nofile;

#ifdef RLIMIT_NOFILE
#define X   RLIMIT_NOFILE
#else
#define X   RLIMIT_OFILE
#endif

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for current check was fail (%d)", errno);
    }
    else {
        last_nofile = nfile.rlim_cur;
    }

    if(fdcap<0) {
        nfile.rlim_cur = nfile.rlim_max;
    }
    else {
        nfile.rlim_cur = fdcap;
    }

    if(setrlimit(X, &nfile)<0) {
        Error("settrlimit() fail (%d)", errno);
    }

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for verify was fail (%d)", errno);
    }
    else {
        Trace("ToBeMaximumFDCapacity: current to %d (max %d), last %d, specified %d\n",
            nfile.rlim_cur, nfile.rlim_max,
            last_nofile, fdcap);
    }

#undef X

}


extern volatile off_t dir_size_max;


static
char *compile_magic = "This program was compiled at " __DATE__ ", " __TIME__;
int listen_fd;

int
main(int argc, char **argv)
{
    int i;
    int chk;
    int ll;

    pthread_attr_t th_attr_high, th_attr_mid, th_attr_low;
    pthread_attr_t *p_th_attr_high, *p_th_attr_mid, *p_th_attr_low;
    struct sched_param th_param;
    int th_policy_type;

    /***
     *** Initialize
     ***/

    MUTEX_LOCK(&stat_lock);
    boot_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);

#if 0
#if 1
    TraceOn();
    {
        extern char log_file[];
        strcpy(log_file, "kotetu.log");
        Trace("*** log_file '%s'\n", log_file);
    }
#else
    TraceOff();
#endif
    InitLogger("kotetuserv");
    Log("START");
    LogSync();
#endif

    SetSignalHandler(SIGPIPE, SIG_IGN);
#if 0
    SetSignalHandler(SIGABRT, SIG_DFL);
    SetSignalHandler(SIGSEGV, SIG_IGN);
#endif

#if 0
    sigemptyset(&handler_sig_mask);
    sigaddset(&handler_sig_mask, SIGTERM);
#if 0
    sigaddset(&handler_sig_mask, SIGPIPE);
#endif
    pthread_sigmask(SIG_UNBLOCK, &handler_sig_mask, NULL);
#endif


    /* Configuration */

    SetGlobalDefaults();
    if(getenv(DEFAULT_CONFIG_ENVNAME)!=NULL) {
        strcpy(config_filename, getenv(DEFAULT_CONFIG_ENVNAME));
    }
    if(config_filename[0]) {
        ReadConfigFile(config_filename);
    }
    ReadCommandlineOptions(argc, argv);

    FixGlobals();


    InitLogger("kotetuserv");
    Log("START %s %s %s ", KOTETU_VER,  __DATE__, __TIME__);
    LogSync();

#if 0
    if(trace_flag) {
        ShowGlobals(stdout);
    }
#endif
    if(trace_flag) {
        TraceOn();
    }
    else {
        TraceOff();
    }

    date_zone_offset_init();

#ifndef NDEBUG
    Trace("not set NDEBUG, so it is debug mode\n\
If you want an actual performance, you should re-compile with -DNDEBUG\n");
#endif
#if 1
    if(getenv("PWD")) {
        Trace("PWD is %s\n", getenv("PWD"));
    }
#endif

#if 1
    Trace("parameters:\n\
    L_URL       %6d\n\
    L_HOST      %6d\n\
    L_PATH      %6d\n\
    L_BUF       %6d\n\
    L_REQBUF    %6d\n", L_URL, L_HOST, L_PATH, L_BUF, L_REQBUF);

    Trace("size of structures:\n\
    skiplist cell   %6d\n\
    skiplist        %6d\n\
    entry           %6d\n\
    frag            %6d\n\
    bucket          %6d\n",
        sizeof(skiplist_cell_t), sizeof(skiplist_t),
        sizeof(skcache_entry_t), sizeof(skcache_frag_t),
        sizeof(bucket_t));

#endif

    ToBeMaximumFDCapacity(-1);

    /* sessions and queues */
#if 0
    InitSession(think_queue_length);
#endif


#if 1
    think_queue_length = require_sess_length;
    InitSession(require_sess_length*(2+require_n_msweeper)+10);

    sq_init(&X_think_queue,     think_queue_length);
    think_queue = &X_think_queue;

    for(i=0;i<require_n_msweeper;i++) {
        sq_init(&X_swapout_queue[i],    think_queue_length);
        swapout_queue[i] = &X_swapout_queue[i];
    }

    sq_init(&X_wait_queue,      think_queue_length);
    wait_queue = &X_wait_queue;

    sq_init(&X_msweep_queue,    2);
    msweep_queue = &X_msweep_queue;

    sq_init(&X_swapin_queue,    2);
    swapin_queue = &X_swapin_queue;
#endif

#if 0
    think_queue_length = require_sess_length;
    InitSession(require_sess_length*4+10);

    sq_init(&X_think_queue,     think_queue_length);
    think_queue = &X_think_queue;

    sq_init(&X_swapout_queue,   think_queue_length*3);
    swapout_queue = &X_swapout_queue;


    sq_init(&X_wait_queue,      2);
    wait_queue = &X_wait_queue;

    sq_init(&X_msweep_queue,    2);
    msweep_queue = &X_msweep_queue;

    sq_init(&X_swapin_queue,    2);
    swapin_queue = &X_swapin_queue;

#endif

#if 0
    think_queue_length = require_sess_length;
    InitSession(require_sess_length*2+6);

    sq_init(&X_think_queue, think_queue_length);
    think_queue = &X_think_queue;

    sq_init(&X_wait_queue,  think_queue_length/4);
    wait_queue = &X_wait_queue;

    sq_init(&X_msweep_queue,    think_queue_length/4);
    msweep_queue = &X_msweep_queue;

    sq_init(&X_swapout_queue,   think_queue_length/4);
    swapout_queue = &X_swapout_queue;

    sq_init(&X_swapin_queue,    think_queue_length/4);
    swapin_queue = &X_swapin_queue;
#endif


#if 1
    Trace("queues:\n\
    think           %p\n\
    wait            %p\n\
    memory sweep    %p\n\
    swapout sweep   %p\n",
        think_queue, wait_queue, msweep_queue, swapout_queue);
#endif

    /* misc. statistically */
    sim_stat_alloc_new(&len_perread_stat, "length of per read(2)",
        K(32),  server_buffer_stat_count);
    sim_stat_U_alloc_new(&object_size_stat,   "object size [KB]",
        1024, K(1),   object_size_stat_count);

    sim_stat_alloc_new(&rcode_stat,    "session response code",
        1000,   rcode_stat_count);

    sim_stat_alloc_new(&url_first_hash_stat,    "first hash value",
        K(2),   url_first_hash_stat_count);
    sim_stat_alloc_new(&url_second_hash_stat,   "second hash value",
        K(2),   url_second_hash_stat_count);
    sim_stat_alloc_new(&url_third_hash_stat,    "third hash value",
        K(2),   url_third_hash_stat_count);
    sim_stat_alloc_new(&url_fourth_hash_stat,   "fourth hash value",
        K(2),   url_fourth_hash_stat_count);

    sim_stat_alloc_new(&req_field_type_stat,    "request field type",
        256,   req_field_type_stat_count);

    sim_stat_alloc_new(&object_alloc_stat,      "object alloc",
        K(10),  object_alloc_stat_count);

    sim_stat_U_alloc_new(&ret_time,
        "session retrieval time [msec]", 100, K(7), ret_time_count);
    sim_stat_U_alloc_new(&ret_time_short,
        "session retrieval time [msec]", 10, K(32), ret_time_short_count);

    sim_stat_U_alloc_new(&mark_time,
        "mark time [msec]", 1, K(10), mark_time_count);
    sim_stat_U_alloc_new(&msweep_time_single,
        "msweep time (single) [msec]", 1, K(10), msweep_time_single_count);
    sim_stat_U_alloc_new(&msweep_time_long,
        "msweep time (long) [msec]", 100, K(10), msweep_time_long_count);

    /* history */
    sim_hist_alloc_new(&comp_hist,              "compaction history",
        30,  comp_hist_area);
    sim_hist_alloc_new(&expire_hist,            "expiration history",
        30,  expire_hist_area);


    /* DNS cache */
    hcache_setup();


#if 0
    /* object cache */
    cache_memory_size = require_total_memory_size - sess_memory_size;
    chk = cache_alloc_area(cache_memory_size,
            require_num_large_block, require_size_large_block,
            require_size_block, require_minimum_size_line);
    if(chk) {
        Error("Sorry, I exit. not enough memory.");
        exit(4);
    }
    if(nblock<0) {
        Error("Sorry, I exit. the number of block is too small (%d).", nblock);
        exit(7);
    }
#endif

#if 1
{
    int bk_num;
    int bk_sk_level, bk_sk_len;
    int frag_size, frag_freelist_len;
    long long U;

    long long skcache_calc_size(
        int *xbk_num, int *xsk_level, int *xsk_len,
        int *xffl_num, int *xffl_size);

    bk_num = require_num_bucket;
    bk_sk_level = require_level_bucket_skiplist;
    frag_size = require_size_frag;

    skcache_calc_param(
        (long long)(require_total_memory_size-sess_memory_size),
        &bk_num, &bk_sk_level, &frag_size, (double)require_ratio_frag_length,
        &bk_sk_len, &frag_freelist_len);

    U = skcache_calc_size(&bk_num, &bk_sk_level, &bk_sk_len, &frag_freelist_len, &frag_size);

#if 1
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

#if 1
    Trace("@ %d %4d %2d %5d %.2f -> %9d %9d => %9d %9d\n",
        boot_date, bk_num, bk_sk_level, frag_size, (double)require_ratio_frag_length,
        bk_sk_len, frag_freelist_len,
        bk_sk_len * bk_num,  frag_freelist_len* bk_num);
#endif


    U = skcache_init(boot_date-86400*7,
            bk_num, bk_sk_level, bk_sk_len, frag_freelist_len, frag_size);
#if 1
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

    cache_memory_size = U;
}
#endif


#if 0
    Trace("Memory allocation:\n\
\ttotal        %12lld // %5ldM+%06ld\n\
\tsession      %12lld // %5ldM+%06ld %6.2f%%\n\
\tobject cache %12lld // %5ldM+%06ld %6.2f%%\n",
        (long long)require_total_memory_size,
          (long)dM(require_total_memory_size),
          (long)mM(require_total_memory_size),
        (long long)sess_memory_size,
          (long)dM(sess_memory_size), (long)mM(sess_memory_size),
          100.0*sess_memory_size/require_total_memory_size,
        (long long)cache_memory_size,
          (long)dM(cache_memory_size), (long)mM(cache_memory_size),
          100.0*cache_memory_size/require_total_memory_size);
#endif

#if 0
    mlist_show();
#endif

    nfrag_obj       = require_nfrag_obj;
    nfrag_dwrite    = require_nfrag_dwrite;
    nfrag_dread     = require_nfrag_dread;
    nfrag_nwrite    = require_nfrag_nwrite;
    nfrag_nread     = require_nfrag_nread;

#if 1
    Trace("I/O information:\n\
  Frags:\n\
    the max number of frags\n\
      (per object)\n\
      object           %3d (%7ldB %6.1fKB) // hard %3d\n\
      (per operation)\n\
      network write    %3d (%7ldB %6.1fKB) // hard %3d\n\
      network read     %3d (%7ldB %6.1fKB) // hard %3d\n\
      disk write       %3d (%7ldB %6.1fKB) // hard %3d\n\
      disk read        %3d (%7ldB %6.1fKB) // hard %3d\n",
        nfrag_obj, (long)nfrag_obj*require_size_frag,
        fdK(nfrag_obj*require_size_frag), HARD_NFRAG_OBJ,
        nfrag_nwrite, (long)nfrag_nwrite*require_size_frag,
        fdK(nfrag_nwrite*require_size_frag), HARD_NFRAG_NWRITE,
        nfrag_nread, (long)nfrag_nread*require_size_frag,
        fdK(nfrag_nread*require_size_frag), HARD_NFRAG_NREAD,
        nfrag_dwrite, (long)nfrag_dwrite*require_size_frag,
        fdK(nfrag_dwrite*require_size_frag), HARD_NFRAG_DWRITE,
        nfrag_dread, (long)nfrag_dread*require_size_frag,
        fdK(nfrag_dread*require_size_frag), HARD_NFRAG_DREAD
    );

#endif



#if 0
    cache_list_entry(stderr);
#endif

#if 0
    {
        struct timeval now;
        gettimeofday(&now, NULL);
        client_scramble_key = now.tv_usec;
    }
#endif
#if 1
        client_scramble_key = 0xabba;
#endif


    /***
     *** Thread launching
     ***/

#ifdef DONT_USE_THREAD_ATTR
    p_th_attr_high  = NULL;
    p_th_attr_mid   = NULL;
    p_th_attr_low   = NULL;

#else
    chk = pthread_attr_init(&th_attr_high);
    chk = pthread_attr_init(&th_attr_mid);
    chk = pthread_attr_init(&th_attr_low);
    Trace("ret pthread_attr_init() %d (%d)\n", chk, errno);

    chk = pthread_attr_getschedparam(&th_attr_high, &th_param);
    Trace("default (initialized) priority value %d\n",
        th_param.sched_priority);
    th_param.sched_priority += th_param.sched_priority; /* double */
    chk = pthread_attr_setschedparam(&th_attr_high, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_mid, &th_param);
    th_param.sched_priority = th_param.sched_priority; /* thru */
    chk = pthread_attr_setschedparam(&th_attr_mid, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_low, &th_param);
    th_param.sched_priority = th_param.sched_priority/2; /* half */
    chk = pthread_attr_setschedparam(&th_attr_low, &th_param);

    chk = pthread_attr_getschedpolicy(&th_attr_mid, &th_policy_type);
    Trace("ret pthread_attr_getschedpolicy() %d (%d)\n", chk, errno);
    Trace("default value %d\n", th_policy_type);

#if 1
    th_policy_type = SCHED_FIFO;
    chk = pthread_attr_setschedpolicy(&th_attr_high, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_mid, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_low, th_policy_type);
#endif

    p_th_attr_high  = &th_attr_high;
    p_th_attr_mid   = &th_attr_mid;
    p_th_attr_low   = &th_attr_low;
#endif





#ifdef TRACE_THREADS
    Trace("launching time is over\n");
#endif

    MUTEX_LOCK(&stat_lock);
    service_start_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);


    /***
     ***
     ***/

    atexit(LogSync);

    MUTEX_LOCK(&stat_lock);
    Trace("initalize take %d sec(s)\n", time(NULL) - boot_date);
    MUTEX_UNLOCK(&stat_lock);

    MUTEX_LOCK(&stat_lock);
    Trace("disk_bench: clear and set very large number to swapout token\n");
    swapout_token_init = (1L<<30);
    MUTEX_UNLOCK(&stat_lock);

    chk = pthread_create(&th_monitor, p_th_attr_mid,
            DiskJudge, NULL);


    /***
     ***
     ***/
    sleep(0);

    /* for child reaping, this thread should be awake at sometime */
    while(1) {
        struct timespec w;

        MUTEX_LOCK(&stat_lock);
        swapout_token = swapout_token_init;
        MUTEX_UNLOCK(&stat_lock);

        msleep(swapout_token_interval);
    }
}

