/*
 * Skiplist & its Buckets
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 *  $Id$
 *
 */

/*
 * skcache.h - object cache procedures for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 *
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * Original algorithm was described by Pascal like langauage.
 * Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

typedef struct _skiplist_cell_t {
    void*                   entity;
    struct _skiplist_cell_t **skip;
} skiplist_cell_t;

typedef struct _skiplist_t {
    int             maxlevel;
    int             level;
    skiplist_cell_t *header;
    skiplist_cell_t *term;
    int             num;
} skiplist_t;


#ifndef SKL_MAXLEVEL
#define SKL_MAXLEVEL    (64)
#endif

#ifndef RADIX_H1_MAX
#define RADIX_H1_MAX    (521)
#endif
#ifndef RADIX_H2_MAX
#define RADIX_H2_MAX    (1031)
#endif
#ifndef RADIX_H3_MAX
#define RADIX_H3_MAX    (1031)
#endif


#ifndef MAGIC_0
#define MAGIC_0     'K'
#endif
#ifndef MAGIC_1
#define MAGIC_1     'w'
#endif

#define XSK_ENTRY   1
#define XSK_FRAG    2

#define XSK_DISKFULL (-27)


#define N_BUCKET_MAX    RADIX_H1_MAX

typedef struct _skcache_frag_t {
    int size;
    int len;
    char *ent;
    struct _skcache_frag_t *next;
} skcache_frag_t;

typedef struct _skcache_entry_t {
    char magic[2];

    int h1, h2, h3, h4;             /* quad hash */
    int oid;
    unsigned char name[L_URL];

    int count;
    int ref;
    int swapout_candidate;
/*
    int owner;
*/
    void *owner;
    int cracked;

    time_t issue_time;
    time_t hot;
    time_t lmd;                     /* last modified */
    time_t expire;                  /* expire */

    off_t totallength;
    skcache_frag_t *m_recorded;
    int d_recorded;

    struct _skcache_entry_t *next;

} skcache_entry_t;


#ifndef MAX_GC_NTARGET
#define MAX_GC_NTARGET (128)
#endif

typedef struct {
    int oid;
    int hot;
    skcache_entry_t *pos;
} skcache_gc_target_t;

typedef struct _bucket_t {
    pthread_mutex_t mutex;

    int             bid;
    int             own;

    skiplist_t      cup;
    int             elemnum;
    int             elemsize;
    int             totalsize;
    skiplist_cell_t *elem_freelist;

    time_t          gc_last;
    int             gc_push;
    int             gc_rest;
    int             gc_checkpoint;
    int             gc_ntarget;
    skcache_gc_target_t gc_targets[MAX_GC_NTARGET];

    int             inmemory;
    int             indisk;
    int             op[4];

    int             dirty2;
    int             dirty3;
    int             dirpos;

    skcache_entry_t *entry_freelist;
    int             entry_use, entry_free, entry_all;
    int             entry_hhwater;
    int             entry_hwater;
    int             entry_lwater;

    int             frag_use;

#ifndef NDEBUG
    skiplist_cell_t *cell_first;
    skiplist_cell_t *cell_last;
#endif
#ifndef NDEBUG
    skcache_entry_t *entry_first;
    skcache_entry_t *entry_last;
#endif
} bucket_t;

#if 0
skiplist_cell_t *z_elem_freelist;
skcache_entry_t *z_entry_freelist;
int entry_use, entry_free;
#endif

extern pthread_mutex_t frag_lock;
extern skcache_frag_t *z_frag_freelist;
extern int frag_use, frag_free;
extern int frag_hhwater, frag_hwater, frag_lwater;

extern double skcache_fratio_hwater;
extern double skcache_fratio_lwater;

#ifdef DEBUG
    void *entity_first;
    void *entity_last;
#endif

#if 0
extern sim_hist_rec fratio_hist;
#endif



/* skiplist */
void sk_rawdump (FILE *fp ,skiplist_t *list ,char *(fn )(void *));
void sk_dump (skiplist_t *list ,FILE *fp ,char *(fn )(void *));
void *sk_find (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
void *sk_assign (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int *st ,int opt );
int sk_insert (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int (ovw )(void **,void *),int opt );
void *sk_delete (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void*),int (rem )(void *),int opt );
int sk_mapdelete (skiplist_t *list ,skiplist_cell_t **flist ,int (ack )(void *,void*),void *);
skiplist_t *sk_new (int nlevel ,void *user_term );

/* bucket */
void bk_dump (bucket_t *bk ,skiplist_t *list ,FILE *fp ,char *(fn )(void *));
skiplist_cell_t *sk_find_listpos (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
int bk_new (bucket_t *bk ,int nlevel ,int nelement ,void *user_term );
int bucket_show (bucket_t *bk );


extern skcache_entry_t *NIL;
extern int skcache_bucket_num;
extern bucket_t *skcache_bucket_base;
extern skcache_entry_t **skcache_entry_freelist_base;
extern skcache_frag_t   **skcache_frag_freelist_base;

int skcache_entry_cmp(void *px, void *py);
char* skcache_label_hash(void *pp);
char* skcache_label_ref(void *pp);
char* skcache_label_short(void *pp);

int skcache_have(int , int , int , int , char *, int , time_t );

long long skcache_init(time_t, int xbucket_num, int xsk_level, int xsk_num,
        int xfl_num, int xfl_size
);


int skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e);
int skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e);

int skcache_dicision_w_th(bucket_t *bk, skcache_entry_t *p, int force,
    time_t now, time_t min, time_t th, time_t max);
int skcache_dicision(bucket_t *bk, skcache_entry_t *p, int force);


skcache_entry_t* skcache_pull_new_entry(bucket_t *bk);
int skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen);

skcache_frag_t* skcache_pull_new_frag(bucket_t *bk);
int skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);

int skcache_update_keeptime();


extern int radix_h1, radix_h2, radix_h3;


/***
 *** directory signature;  variables and functions
 ***/

extern pthread_mutex_t dir_mutex;
extern int dir_num;
extern int *dir_lock;
extern off_t *dir_sz;
extern int *dir_count;
extern int *dir_time;
extern volatile off_t dir_size_hwater, dir_size_lwater;

extern int skcache_signature_refresh;

#ifdef USE_3LEVEL_DIR
#define DSIG_POS(x1, x2, x3)    (x1*radix_h2*radix_h3 + x2*radix_h3 + x3)
#else
#define DSIG_POS(x1, x2, x3)    (x1*radix_h2 + x2)
#endif


#ifdef NOT_HAVE_UINT
#define u_int32_t    unsigned int
#endif

#ifndef SIZE_OF_UINT
#ifdef SIZE_OF_INT
#define SIZE_OF_UINT    SIZE_OF_INT
#else
#define SIZE_OF_UINT    (4)
#endif
#endif


#define DSIG_SIG_MAX        -1  /* not defined, it was solved automatically */
/*
#define DSIG_BIT_WIDTH      (1024)
*/
#define DSIG_BIT_WIDTH      (7367)
#define DSIG_VALUE_RANGE    (8000)
#define DSIG_VALUE_RES      (DSIG_VALUE_RANGE/DSIG_BIT_WIDTH)
/*
#define DSIG_ELEM_WIDTH     (DSIG_BIT_WIDTH/(8*SIZE_OF_UINT))
*/
#define DSIG_ELEM_WIDTH  ((DSIG_BIT_WIDTH+(8*SIZE_OF_UINT-1))/(8*SIZE_OF_UINT))
#define DSIG_BYTE_WIDTH     (DSIG_ELEM_WIDTH*SIZE_OF_UINT)

void dsig_mkhumanstr(u_int32_t *s, char *line);
void dsig_show(u_int32_t *s);
void dsig_set(u_int32_t *s, int i);
int dsig_isset(u_int32_t *s, int i);
int dsig_allset(u_int32_t *s);
int dsig_allclear(u_int32_t *s);


extern u_int32_t *dir_sig;

#define DSIG_SET(n,i)   \
    (dir_sig[n*DSIG_ELEM_WIDTH+i/(SIZE_OF_UINT*8)] \
        |= (1UL << i%(SIZE_OF_UINT*8)))
#define DSIG_ISSET(n,i) \
    ((dir_sig[n*DSIG_ELEM_WIDTH+i/(SIZE_OF_UINT*8)] \
        & (1UL << i%(SIZE_OF_UINT*8))) ? 1 : 0)

#define DSIG_TMP_SET(v,i)   \
    (v[i/(SIZE_OF_UINT*8)] \
        |= (1UL << i%(SIZE_OF_UINT*8)))
#define DSIG_TMP_ISSET(v,i) \
    ((v[i/(SIZE_OF_UINT*8)] \
        & (1UL << i%(SIZE_OF_UINT*8))) ? 1 : 0)

#define DSIG_ALLCLEAR(n)    dsig_allclear(&dir_sig[n*DSIG_ELEM_WIDTH])
#define DSIG_ALLSET(n)      dsig_allset(&dir_sig[n*DSIG_ELEM_WIDTH])


#ifndef DSIG_MIN_GAP
#define DSIG_MIN_GAP    (5)
#endif

#ifndef DSIG_REFRESH
#define DSIG_REFRESH    (3600)
#endif


/***
 ***
 ***/

#ifdef CHECK_BKMUTEX

#define LOCK_BK(x)  if(x) {\
    Trace("bkl do   th %08x bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    pthread_mutex_lock(&(x->mutex)); \
    Trace("bkl done th %08x bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    }\
    else {\
    Trace("bkl null th %08x bk. *strange*\n", pthread_self()); \
    }

#define TRYLOCK_BK(x,r)  if(x) {\
    Trace("bkt do   th %08x bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    r = pthread_mutex_trylock(&(x->mutex)); \
    Trace("bkt done th %08x bk# %d, : %s %5d : ret %d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__, r); \
    }\
    else {\
    Trace("bkt null th %08x bk. *strange*\n", pthread_self()); \
    }

#define UNLOCK_BK(x)    if(x) {int r;\
    Trace("bku do   th %08x bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    r = pthread_mutex_unlock(&(x->mutex)); \
    Trace("bku done th %08x bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    }\
    else {\
    Trace("bku null th %08x bk. *strange*\n", pthread_self()); \
    }

#else

#define LOCK_BK(x)          pthread_mutex_lock(&(x->mutex))
#define TRYLOCK_BK(x,r)     r = pthread_mutex_trylock(&(x->mutex))
#define UNLOCK_BK(x)        pthread_mutex_unlock(&(x->mutex))

#endif




#ifndef HARD_NFRAG_OBJ
#define HARD_NFRAG_OBJ      (1024)
#endif
#ifndef HARD_NFRAG_NWRITE
#define HARD_NFRAG_NWRITE   (12)
#endif
#ifndef HARD_NFRAG_NREAD
#define HARD_NFRAG_NREAD    (12)
#endif
#ifndef HARD_NFRAG_DWRITE
#define HARD_NFRAG_DWRITE   (12)
#endif
#ifndef HARD_NFRAG_DREAD
#define HARD_NFRAG_DREAD    (12)
#endif
         
extern int nfrag_obj;
extern int nfrag_nwrite;
extern int nfrag_nread;
extern int nfrag_dwrite;
extern int nfrag_dread;

    

